-- Little People Crowd Simulation with Faux Walk Animation
-- Touch the screen to make the crowd rush to that point.
-- When no touch is active, each little person wanders around.
-- Each person shows an abstract walking animation:
-- two crossing lines that oscillate between overlapping and a narrow X shape.

local people = {}      -- table to store our little people
local numPeople = 300  -- number of people in the simulation
local touchPos = nil   -- current touch position (nil when no touch)
-- Custom map function
function map(value, low1, high1, low2, high2)
    return low2 + (value - low1) * (high2 - low2) / (high1 - low1)
end

function setup()
    -- Set rounded line caps for a smoother limb appearance.
    lineCapMode(ROUND)
    
    -- Initialize each little person with a random starting position,
    -- velocity, and random walk parameters (phase, frequency, amplitude)
    for i = 1, numPeople do
        local person = {
            x = math.random(WIDTH),
            y = math.random(HEIGHT),
            vx = 0,
            vy = 0,
            phase = math.random() * 2 * math.pi,          -- unique phase offset
            walkFreq = 1 + math.random() * 2,               -- frequency between 1 and 3
            walkAmplitude = 2 + math.random() * 4           -- amplitude between 2 and 6
        }
        person.phaseTop = math.random() * 2 * math.pi
        person.phaseBottom = math.random() * 2 * math.pi
        person.phaseTop2 = math.random() * 2 * math.pi
        person.phaseBottom2 = math.random() * 2 * math.pi
        person.angleFreq = 1.5      -- Frequency (cycles per second)
        person.angleRange = 80      -- Maximum angular variation in degrees
        person.crowdPhase = math.random() * 2 * math.pi
        table.insert(people, person)
    end
end

function touched(touch)
    -- Record the touch position when the user touches or moves a finger,
    -- and clear it when the touch ends.
    if touch.state == BEGAN or touch.state == MOVING then
        touchPos = { x = touch.x, y = touch.y }
    else
        touchPos = nil
    end
end

function draw()
    background(40, 40, 50)
    
    for i, person in ipairs(people) do
        -- (Use the modified touch code above)
        if touchPos then
            local crowdRange = 20          -- Maximum offset in pixels
            local crowdSpeed = 1           -- Speed factor for dynamic offset variation
            local offsetX = map(noise(person.crowdPhase + ElapsedTime * crowdSpeed), 0, 1, -crowdRange, crowdRange)
            local offsetY = map(noise(person.crowdPhase + 100 + ElapsedTime * crowdSpeed), 0, 1, -crowdRange, crowdRange)
            local targetX = touchPos.x + offsetX
            local targetY = touchPos.y + offsetY
            local dx = targetX - person.x
            local dy = targetY - person.y
            local dist = math.sqrt(dx * dx + dy * dy)
            if dist > 1 then
                local accel = 500
                local ax = (dx / dist) * accel
                local ay = (dy / dist) * accel
                person.vx = person.vx + ax * DeltaTime
                person.vy = person.vy + ay * DeltaTime
            end
        else
            local wanderAccel = 100
            local ax = (math.random() - 0.5) * wanderAccel
            local ay = (math.random() - 0.5) * wanderAccel
            person.vx = person.vx + ax * DeltaTime
            person.vy = person.vy + ay * DeltaTime
        end
        
        person.vx = person.vx * 0.95
        person.vy = person.vy * 0.95
        person.x = person.x + person.vx * DeltaTime
        person.y = person.y + person.vy * DeltaTime
        
        drawDoodleWalker(person)
    end
end

function drawDoodleWalker(person)
    pushMatrix()
    translate(person.x, person.y)
    
    lineCapMode(ROUND)
    stroke(255)
    strokeWidth(2)
    
    local lineLength = 6
    
    -- Compute the base orientation from movement using math.atan.
    local speed = math.sqrt(person.vx * person.vx + person.vy * person.vy)
    local baseAngle = speed > 0.1 and math.atan(person.vy, person.vx) or 0
    
    local freq = person.angleFreq or 1.5
    local range = person.angleRange or 15  -- maximum variation in degrees
    
    -- Scale the oscillation frequency based on movement speed.
    local dynamicFreq = freq * (1 + speed / 350)
    local t = ElapsedTime
    
    -- First pair of segments:
    local topOffset1 = math.sin(t * dynamicFreq + (person.phaseTop or 0)) * range
    local bottomOffset1 = math.sin(t * dynamicFreq + (person.phaseBottom or 0)) * range
    
    local effectiveAngleTop1 = baseAngle + topOffset1
    local effectiveAngleBottom1 = baseAngle + 180 + bottomOffset1
    
    local topX1 = math.cos(math.rad(effectiveAngleTop1)) * (lineLength/2)
    local topY1 = math.sin(math.rad(effectiveAngleTop1)) * (lineLength/2)
    local bottomX1 = math.cos(math.rad(effectiveAngleBottom1)) * (lineLength/2)
    local bottomY1 = math.sin(math.rad(effectiveAngleBottom1)) * (lineLength/2)
    
    line(0, 0, topX1, topY1)
    line(0, 0, bottomX1, bottomY1)
    
    -- Second pair of segments (duplicating the same algorithm but with independent phase values):
    local topOffset2 = math.sin(t * dynamicFreq + (person.phaseTop2 or 0)) * range
    local bottomOffset2 = math.sin(t * dynamicFreq + (person.phaseBottom2 or 0)) * range
    
    local effectiveAngleTop2 = baseAngle + topOffset2
    local effectiveAngleBottom2 = baseAngle + 180 + bottomOffset2
    
    local topX2 = math.cos(math.rad(effectiveAngleTop2)) * (lineLength/2)
    local topY2 = math.sin(math.rad(effectiveAngleTop2)) * (lineLength/2)
    local bottomX2 = math.cos(math.rad(effectiveAngleBottom2)) * (lineLength/2)
    local bottomY2 = math.sin(math.rad(effectiveAngleBottom2)) * (lineLength/2)
    
    line(0, 0, topX2, topY2)
    line(0, 0, bottomX2, bottomY2)
    
    -- Optionally, draw a small central body.
    fill(255)
    noStroke()
    --ellipse(0, 0, 8)
    
    popMatrix()
end