-- Little People Crowd Simulation with Faux Walk Animation
-- Touch the screen to make the crowd rush to that point.
-- When no touch is active, each little person wanders around.
-- Each person shows an abstract walking animation:
-- two crossing lines that oscillate between overlapping and a narrow X shape.

local people = {}      -- table to store our little people
local numPeople = 300  -- number of people in the simulation
local touchPos = nil   -- current touch position (nil when no touch)

function setup()
    -- Set rounded line caps for a smoother limb appearance.
    lineCapMode(ROUND)
    
    -- Initialize each little person with a random starting position,
    -- velocity, and random walk parameters (phase, frequency, amplitude)
    for i = 1, numPeople do
        local person = {
            x = math.random(WIDTH),
            y = math.random(HEIGHT),
            vx = 0,
            vy = 0,
            phase = math.random() * 2 * math.pi,          -- unique phase offset
            walkFreq = 1 + math.random() * 2,               -- frequency between 1 and 3
            walkAmplitude = 2 + math.random() * 4           -- amplitude between 2 and 6
        }
        person.phaseTop = math.random() * 2 * math.pi
        person.phaseBottom = math.random() * 2 * math.pi
        person.phaseTop2 = math.random() * 2 * math.pi
        person.phaseBottom2 = math.random() * 2 * math.pi
        person.angleFreq = 1.5      -- Frequency (cycles per second)
        person.angleRange = 80      -- Maximum angular variation in degrees
        person.crowdPhase = math.random() * 2 * math.pi
        person.reactionFactor = 1 + ((math.random() - 0.5)^3 * 0.8)
        person.wanderFactor   = 1 + ((math.random() - 0.5)^3 * 0.8)
        table.insert(people, person)
    end
end

function touched(touch)
    if touch.state == BEGAN or touch.state == MOVING then
        touchPos = { x = touch.x, y = touch.y }
        lastTouchTime = ElapsedTime
    else
        touchPos = nil
    end
end
-- Global variable to track the last time a touch was active.
lastTouchTime = 0

-- Helper: Custom map function.
function map(value, low1, high1, low2, high2)
    return low2 + (value - low1) * (high2 - low2) / (high1 - low1)
end

-- Jitter helper functions.
function jitterAngle(phase, speed)
    return map(noise(phase + ElapsedTime * speed), 0, 1, 0, 2 * math.pi)
end

function jitterRadius(phase, speed, range)
    return map(noise(phase + 100 + ElapsedTime * speed), 0, 1, -range, range)
end

-- Compute the dynamic crowd target (when touch is active).
function calcCrowdTarget(person)
    local ringRadius = 60         -- desired ring radius around the touch
    local jitterRange = 10        -- maximum radial jitter
    local crowdSpeed = 1          -- speed factor for dynamic offset variation
    
    local angle = jitterAngle(person.crowdPhase, crowdSpeed)
    local radiusOffset = jitterRadius(person.crowdPhase, crowdSpeed, jitterRange)
    local targetRadius = ringRadius + radiusOffset
    
    local targetX = touchPos.x + targetRadius * math.cos(angle)
    local targetY = touchPos.y + targetRadius * math.sin(angle)
    return targetX, targetY
end

-- Compute a random target anywhere on screen.
function calcRandomTarget()
    local targetX = math.random(WIDTH)
    local targetY = math.random(HEIGHT)
    return targetX, targetY
end

-- Apply repulsion forces from all other persons.
-- Now, repulsion is only applied when touch is active.
function applyRepulsionFor(person, index)
    if not touchPos then
        return
    end
    local minDistance = 10       -- minimum separation
    local repulsionFactor = 50   -- repulsion strength
    local repulsionX, repulsionY = 0, 0
    for j, other in ipairs(people) do
        if j ~= index then
            local dx = person.x - other.x
            local dy = person.y - other.y
            local d = math.sqrt(dx * dx + dy * dy)
            if d < minDistance and d > 0 then
                local force = repulsionFactor * (minDistance - d) / d
                repulsionX = repulsionX + dx * force
                repulsionY = repulsionY + dy * force
            end
        end
    end
    person.vx = person.vx + repulsionX * DeltaTime
    person.vy = person.vy + repulsionY * DeltaTime
end

-- Update a single person.
function updatePerson(person, index)
    if touchPos then
        -- When touch is active, follow the dynamic crowd target.
        person.randomImpulseApplied = false  -- reset impulse flag
        local targetX, targetY = calcCrowdTarget(person)
        local dx = targetX - person.x
        local dy = targetY - person.y
        local dist = math.sqrt(dx * dx + dy * dy)
        if dist > 1 then
            local reactionFactor = person.reactionFactor or 1.0
            local accel = 500 * reactionFactor
            local ax = (dx / dist) * accel
            local ay = (dy / dist) * accel
            person.vx = person.vx + ax * DeltaTime
            person.vy = person.vy + ay * DeltaTime
        end
        lastTouchTime = ElapsedTime  -- update last touch time
    else
        -- When no touch is active, apply a one-time random impulse toward a random target.
        if not person.randomImpulseApplied then
            local targetX, targetY = calcRandomTarget()
            person.randomImpulseApplied = true
            local dx = targetX - person.x
            local dy = targetY - person.y
            local dist = math.sqrt(dx * dx + dy * dy)
            if dist > 1 then
                local fraction = math.random() * 0.5 + 0.5  -- random fraction in [0.5, 1.0]
                local impulseMagnitude = dist * fraction
                local impulseAngle = math.atan(dy, dx)
                person.vx = person.vx + impulseMagnitude * math.cos(impulseAngle)
                person.vy = person.vy + impulseMagnitude * math.sin(impulseAngle)
            end
        end
    end
    
    -- Apply repulsion only when touch is active.
    applyRepulsionFor(person, index)
    
    -- Apply damping.
    person.vx = person.vx * 0.95
    person.vy = person.vy * 0.95
    
    -- Update position.
    person.x = person.x + person.vx * DeltaTime
    person.y = person.y + person.vy * DeltaTime
end

-- Main draw loop.
function draw()
    background(40, 40, 50)
    
    for i, person in ipairs(people) do
        updatePerson(person, i)
        drawDoodleWalker(person)
    end
    
    wasTouching = (touchPos ~= nil)
end

function drawDoodleWalker(person)
    pushMatrix()
    translate(person.x, person.y)
    
    lineCapMode(ROUND)
    stroke(255)
    strokeWidth(2)
    
    local lineLength = 6
    
    -- Compute the base orientation from movement using math.atan.
    local speed = math.sqrt(person.vx * person.vx + person.vy * person.vy)
    local baseAngle = speed > 0.1 and math.atan(person.vy, person.vx) or 0
    
    local freq = person.angleFreq or 1.5
    local range = person.angleRange or 15  -- maximum variation in degrees
    
    -- Scale the oscillation frequency based on movement speed.
    local dynamicFreq = freq * (1 + speed / 350)
    local t = ElapsedTime
    
    -- First pair of segments:
    local topOffset1 = math.sin(t * dynamicFreq + (person.phaseTop or 0)) * range
    local bottomOffset1 = math.sin(t * dynamicFreq + (person.phaseBottom or 0)) * range
    
    local effectiveAngleTop1 = baseAngle + topOffset1
    local effectiveAngleBottom1 = baseAngle + 180 + bottomOffset1
    
    local topX1 = math.cos(math.rad(effectiveAngleTop1)) * (lineLength/2)
    local topY1 = math.sin(math.rad(effectiveAngleTop1)) * (lineLength/2)
    local bottomX1 = math.cos(math.rad(effectiveAngleBottom1)) * (lineLength/2)
    local bottomY1 = math.sin(math.rad(effectiveAngleBottom1)) * (lineLength/2)
    
    line(0, 0, topX1, topY1)
    line(0, 0, bottomX1, bottomY1)
    
    -- Second pair of segments (duplicating the same algorithm but with independent phase values):
    local topOffset2 = math.sin(t * dynamicFreq + (person.phaseTop2 or 0)) * range
    local bottomOffset2 = math.sin(t * dynamicFreq + (person.phaseBottom2 or 0)) * range
    
    local effectiveAngleTop2 = baseAngle + topOffset2
    local effectiveAngleBottom2 = baseAngle + 180 + bottomOffset2
    
    local topX2 = math.cos(math.rad(effectiveAngleTop2)) * (lineLength/2)
    local topY2 = math.sin(math.rad(effectiveAngleTop2)) * (lineLength/2)
    local bottomX2 = math.cos(math.rad(effectiveAngleBottom2)) * (lineLength/2)
    local bottomY2 = math.sin(math.rad(effectiveAngleBottom2)) * (lineLength/2)
    
    line(0, 0, topX2, topY2)
    line(0, 0, bottomX2, bottomY2)
    
    -- Optionally, draw a small central body.
    fill(255)
    noStroke()
    --ellipse(0, 0, 8)
    
    popMatrix()
end