-- Full Code: Colorful Crowd Simulation
-- Each person (or “agent”) belongs to one of three pastel color groups.
-- When you touch the screen, if the touch is near an agent,
-- that agent’s color group becomes “active” and only agents of that group converge on a dynamic target
-- (computed using noise-based jitter around a ring centered on the touch).
-- All other agents receive only a small random impulse.

-- Global variables
touchPos = nil              -- Current touch position (nil when no touch)
activeColorGroup = nil      -- Color group determined by the touch (if any)
wasTouching = false         -- Tracks whether there was an active touch in the previous frame
lastTouchTime = 0           -- Time when a touch was last detected
people = {}                 -- List of all agents
fps = 0
lastTime = ElapsedTime

-- Number of people per color group
numPeoplePerGroup = 300
parameter.integer("numPeoplePerGroup", 10, 400, numPeoplePerGroup, function(newVal)
    numPeoplePerGroup = newVal
    resetPeople()  -- Reinitialize the agents with the new group size.
end)

-- Group colors (pastel)
numColorGroups = 7
parameter.integer("numColorGroups", 2, 10, numColorGroups)
groupColors = {}
-- HSL to RGB conversion: h, s, l in [0,1]
function hslToRgb(h, s, l)
    if s == 0 then
        return l*255, l*255, l*255
    else
        local function hue2rgb(p, q, t)
            if t < 0 then t = t + 1 end
            if t > 1 then t = t - 1 end
            if t < 1/6 then return p + (q - p) * 6 * t end
            if t < 1/2 then return q end
            if t < 2/3 then return p + (q - p) * (2/3 - t) * 6 end
            return p
        end
        local q = l < 0.5 and l * (1 + s) or l + s - l * s
        local p = 2 * l - q
        local r = hue2rgb(p, q, h + 1/3)
        local g = hue2rgb(p, q, h)
        local b = hue2rgb(p, q, h - 1/3)
        return r * 255, g * 255, b * 255
    end
end

function updateGroupColors()
    groupColors = {}
    local baseHue = math.random()  -- or use a fixed value if you prefer consistency
    for i = 1, numColorGroups do
        local hue = (baseHue + (i - 1) / numColorGroups) % 1
        local r, g, b = hslToRgb(hue, 0.55, 0.85)
        groupColors[i] = { r = r, g = g, b = b }
    end
end

-- Helper: map function (similar to Processing’s map)
function map(value, low1, high1, low2, high2)
    return low2 + (value - low1) * (high2 - low2) / (high1 - low1)
end

-- Jitter helper functions for dynamic target calculation
function jitterAngle(phase, speed)
    return map(noise(phase + ElapsedTime * speed), 0, 1, 0, 2 * math.pi)
end

function jitterRadius(phase, speed, range)
    return map(noise(phase + 100 + ElapsedTime * speed), 0, 1, -range, range)
end

-- Compute the dynamic crowd target for an agent when a touch is active.
function calcCrowdTarget(person)
    local ringRadius = 60         -- Desired ring radius around the touch
    local jitterRange = 10        -- Maximum radial jitter (in pixels)
    local crowdSpeed = 1          -- Speed factor for noise variation
    local angle = jitterAngle(person.crowdPhase, crowdSpeed)
    local radiusOffset = jitterRadius(person.crowdPhase, crowdSpeed, jitterRange)
    local targetRadius = ringRadius + radiusOffset
    local targetX = touchPos.x + targetRadius * math.cos(angle)
    local targetY = touchPos.y + targetRadius * math.sin(angle)
    return targetX, targetY
end

-- Compute a random target anywhere on screen.
function calcRandomTarget()
    local targetX = math.random(WIDTH)
    local targetY = math.random(HEIGHT)
    return targetX, targetY
end

-- Update a single agent (called once per frame per agent)
function updatePerson(person, index)
    if touchPos and activeColorGroup == person.colorGroup then
        -- For agents in the active color group, follow the dynamic target.
        person.randomImpulseApplied = false  -- reset impulse flag
        local targetX, targetY = calcCrowdTarget(person)
        local dx = targetX - person.x
        local dy = targetY - person.y
        local dist = math.sqrt(dx * dx + dy * dy)
        if dist > 1 then
            local reactionFactor = person.reactionFactor or 1.0
            local accel = 500 * reactionFactor
            local ax = (dx / dist) * accel
            local ay = (dy / dist) * accel
            person.vx = person.vx + ax * DeltaTime
            person.vy = person.vy + ay * DeltaTime
        end
        lastTouchTime = ElapsedTime
    else
        -- For agents not in the active group (or when no touch), apply a one-time small random impulse.
        if not person.randomImpulseApplied then
            local targetX, targetY = calcRandomTarget()
            person.randomImpulseApplied = true
            local dx = targetX - person.x
            local dy = targetY - person.y
            local dist = math.sqrt(dx * dx + dy * dy)
            if dist > 1 then
                local fraction = math.random() * 0.2 + 0.1  -- Random fraction in [0.1, 0.3]
                local impulseMagnitude = dist * fraction
                local impulseAngle = math.atan(dy, dx)
                person.vx = person.vx + impulseMagnitude * math.cos(impulseAngle)
                person.vy = person.vy + impulseMagnitude * math.sin(impulseAngle)
            end
        end
    end
    
    -- Apply damping to smooth motion.
    person.vx = person.vx * 0.95
    person.vy = person.vy * 0.95
    
    -- Update the agent's position.
    person.x = person.x + person.vx * DeltaTime
    person.y = person.y + person.vy * DeltaTime
    
    -- Side walls: Clamp positions and reverse velocity upon hitting an edge.
    if person.x < 0 then 
        person.x = 0 
        person.vx = -person.vx 
    elseif person.x > WIDTH then 
        person.x = WIDTH 
        person.vx = -person.vx 
    end
    if person.y < 0 then 
        person.y = 0 
        person.vy = -person.vy 
    elseif person.y > HEIGHT then 
        person.y = HEIGHT 
        person.vy = -person.vy 
    end
end

-- Draw the "doodle" for each agent.
-- This function uses noise-based oscillation to draw two pairs of crossing lines,
-- and fills the center with the agent's color.
function drawDoodleWalker(person)
    pushMatrix()
    translate(person.x, person.y)
    lineCapMode(ROUND)
    local col = groupColors[person.colorGroup] or {r = 255, g = 255, b = 255}
    stroke(col.r, col.g, col.b)
    strokeWidth(3)
    local lineLength = 7.5
    local speed = math.sqrt(person.vx * person.vx + person.vy * person.vy)
    local baseAngle = speed > 0.1 and math.atan(person.vy, person.vx) or 0
    local freq = person.angleFreq or 1.5
    local rangeAngle = person.angleRange or 80  -- in degrees
    local t = ElapsedTime
    
    -- First pair of segments
    local topOffset1 = math.sin(t * freq + (person.phaseTop or 0)) * rangeAngle
    local bottomOffset1 = math.sin(t * freq + (person.phaseBottom or 0)) * rangeAngle
    local effectiveAngleTop1 = baseAngle + topOffset1
    local effectiveAngleBottom1 = baseAngle + 180 + bottomOffset1
    local topX1 = math.cos(math.rad(effectiveAngleTop1)) * (lineLength / 2)
    local topY1 = math.sin(math.rad(effectiveAngleTop1)) * (lineLength / 2)
    local bottomX1 = math.cos(math.rad(effectiveAngleBottom1)) * (lineLength / 2)
    local bottomY1 = math.sin(math.rad(effectiveAngleBottom1)) * (lineLength / 2)
    line(0, 0, topX1, topY1)
    line(0, 0, bottomX1, bottomY1)
    
    -- Second pair of segments (using independent phase values)
    local topOffset2 = math.sin(t * freq + (person.phaseTop2 or 0)) * rangeAngle
    local bottomOffset2 = math.sin(t * freq + (person.phaseBottom2 or 0)) * rangeAngle
    local effectiveAngleTop2 = baseAngle + topOffset2
    local effectiveAngleBottom2 = baseAngle + 180 + bottomOffset2
    local topX2 = math.cos(math.rad(effectiveAngleTop2)) * (lineLength / 2)
    local topY2 = math.sin(math.rad(effectiveAngleTop2)) * (lineLength / 2)
    local bottomX2 = math.cos(math.rad(effectiveAngleBottom2)) * (lineLength / 2)
    local bottomY2 = math.sin(math.rad(effectiveAngleBottom2)) * (lineLength / 2)
    line(0, 0, topX2, topY2)
    line(0, 0, bottomX2, bottomY2)
    
    -- Draw a small central ellipse in the group color.
    fill(col.r, col.g, col.b)
    noStroke()
    popMatrix()
end

function resetPeople()
    people = {}
    for group = 1, numColorGroups do
        for i = 1, numPeoplePerGroup do
            local person = {}
            person.x = math.random(WIDTH)
            person.y = math.random(HEIGHT)
            person.vx = 0
            person.vy = 0
            person.phaseTop = math.random() * 2 * math.pi
            person.phaseBottom = math.random() * 2 * math.pi
            person.phaseTop2 = math.random() * 2 * math.pi
            person.phaseBottom2 = math.random() * 2 * math.pi
            person.crowdPhase = math.random() * 2 * math.pi
            person.angleFreq = 1.5
            person.angleRange = 80   -- in degrees
            person.reactionFactor = 1 + ((math.random() - 0.5) ^ 3 * 0.8)
            person.wanderFactor = 1 + ((math.random() - 0.5) ^ 3 * 0.8)
            person.randomImpulseApplied = false
            person.colorGroup = group
            table.insert(people, person)
        end
    end
end

function setup()
    fps = 0
    lastTime = ElapsedTime
    resetPeople()
    updateGroupColors()
    prevNumColorGroups = numColorGroups  -- save the initial value
end

-- Draw: update each agent and draw it.
function draw()
    background(40, 40, 50)
    -- Check if the slider changed, and if so, update groups and reset simulation.
    if numColorGroups ~= prevNumColorGroups then
        updateGroupColors()
        resetPeople()
        prevNumColorGroups = numColorGroups
    end
    for i, person in ipairs(people) do
        updatePerson(person, i)
        drawDoodleWalker(person)
    end
    wasTouching = (touchPos ~= nil)
end

-- Touched function:
-- When a touch is active, record the touch position.
-- Also, if the touch is close to an agent, set that agent’s color group as active.
function touched(touch)
    if touch.state == BEGAN or touch.state == MOVING then
        touchPos = { x = touch.x, y = touch.y }
        lastTouchTime = ElapsedTime
        activeColorGroup = nil
        for i, person in ipairs(people) do
            local dx = person.x - touch.x
            local dy = person.y - touch.y
            if math.sqrt(dx * dx + dy * dy) < 20 then
                activeColorGroup = person.colorGroup
                break
            end
        end
    else
        touchPos = nil
        activeColorGroup = nil
    end
end