-- Parameter definitions (set before setup)
maxFlockingDistance = 100
parameter.number("maxFlockingDistance", 50, 200, maxFlockingDistance)
numColorGroups = 3
parameter.number("numColorGroups", 2, 10, numColorGroups)
numPeoplePerGroup = 300
parameter.integer("numPeoplePerGroup", 10, 400, numPeoplePerGroup, function(newVal)
    numPeoplePerGroup = newVal
    resetPeople()
    updateGroupColors()
    resetCorralData()
end)


--------------------------------------------------
-- Global Variables
touchPos = nil              -- Current touch position (nil when no touch)
activeColorGroup = nil      -- Color group determined by the touch (if any)
wasTouching = false         -- Tracks whether there was an active touch in the previous frame
lastTouchTime = 0           -- Time when a touch was last detected
people = {}                 -- List of all agents
fps = 0
lastTime = ElapsedTime

-- Game state and corral target data
gameState = "active"        -- "active" while game is ongoing; "finished" when all groups are corralled
currentTarget = nil         -- Current target circle (if any)
initialCount = {}           -- Number of agents per color group at start
corralledCount = {}         -- Number of agents per color group that have been corralled

-- Group colors (pastel) – will be updated by updateGroupColors()
groupColors = {}

--------------------------------------------------
-- Utility Functions

-- HSL to RGB conversion: h, s, l in [0,1]
function hslToRgb(h, s, l)
    if s == 0 then
        return l*255, l*255, l*255
    else
        local function hue2rgb(p, q, t)
            if t < 0 then t = t + 1 end
            if t > 1 then t = t - 1 end
            if t < 1/6 then return p + (q - p) * 6 * t end
            if t < 1/2 then return q end
            if t < 2/3 then return p + (q - p) * (2/3 - t) * 6 end
            return p
        end
        local q = l < 0.5 and l * (1 + s) or l + s - l * s
        local p = 2 * l - q
        local r = hue2rgb(p, q, h + 1/3)
        local g = hue2rgb(p, q, h)
        local b = hue2rgb(p, q, h - 1/3)
        return r * 255, g * 255, b * 255
    end
end

-- Map function (like Processing’s map)
function map(value, low1, high1, low2, high2)
    return low2 + (value - low1) * (high2 - low2) / (high1 - low1)
end

-- Jitter helper functions for dynamic target calculation
function jitterAngle(phase, speed)
    return map(noise(phase + ElapsedTime * speed), 0, 1, 0, 2 * math.pi)
end

function jitterRadius(phase, speed, range)
    return map(noise(phase + 100 + ElapsedTime * speed), 0, 1, -range, range)
end

--------------------------------------------------
-- Group Colors and Agent Reset Functions

function updateGroupColors()
    groupColors = {}
    local baseHue = math.random()  -- can also use a fixed base for consistency
    for i = 1, numColorGroups do
        local hue = (baseHue + (i - 1) / numColorGroups) % 1
        local r, g, b = hslToRgb(hue, 0.4, 0.85)
        groupColors[i] = { r = r, g = g, b = b }
    end
end

function resetPeople()
    people = {}
    for group = 1, numColorGroups do
        for i = 1, numPeoplePerGroup do
            local person = {}
            person.x = math.random() * WIDTH
            person.y = math.random() * HEIGHT
            person.vx = 0
            person.vy = 0
            person.phaseTop = math.random() * 2 * math.pi
            person.phaseBottom = math.random() * 2 * math.pi
            person.phaseTop2 = math.random() * 2 * math.pi
            person.phaseBottom2 = math.random() * 2 * math.pi
            person.crowdPhase = math.random() * 2 * math.pi
            person.angleFreq = 1.5
            person.angleRange = 80   -- in degrees
            person.reactionFactor = 1 + ((math.random() - 0.5) ^ 3 * 0.8)
            person.wanderFactor = 1 + ((math.random() - 0.5) ^ 3 * 0.8)
            person.randomImpulseApplied = false
            person.colorGroup = group
            person.corralled = false
            table.insert(people, person)
        end
    end
end

function resetCorralData()
    gameState = "active"
    currentTarget = nil
    initialCount = {}
    corralledCount = {}
    for i, person in ipairs(people) do
        local grp = person.colorGroup
        initialCount[grp] = (initialCount[grp] or 0) + 1
        corralledCount[grp] = 0
    end
end

function resetGame()
    updateGroupColors()
    resetPeople()
    resetCorralData()
end

--------------------------------------------------
-- Target (Corral) Functions

-- Compute the dynamic target circle for a flocking group.
function calcCrowdTarget(person)
    local ringRadius = 60         -- Desired ring radius around the touch
    local jitterRange = 10        -- Maximum radial jitter (in pixels)
    local crowdSpeed = 1          -- Speed factor for noise variation
    local angle = jitterAngle(person.crowdPhase, crowdSpeed)
    local radiusOffset = jitterRadius(person.crowdPhase, crowdSpeed, jitterRange)
    local targetRadius = ringRadius + radiusOffset
    local targetX = touchPos.x + targetRadius * math.cos(angle)
    local targetY = touchPos.y + targetRadius * math.sin(angle)
    return targetX, targetY
end

-- Compute a random target (used for non-flocking agents when touch is inactive)
function calcRandomTarget()
    local targetX = math.random() * WIDTH
    local targetY = math.random() * HEIGHT
    return targetX, targetY
end

-- Update the current target circle:
function updateTarget()
    if gameState ~= "active" then return end
    if not currentTarget then
        local availableGroups = {}
        for grp = 1, numColorGroups do
            if (corralledCount[grp] or 0) < (initialCount[grp] or 1) then
                table.insert(availableGroups, grp)
            end
        end
        if #availableGroups > 0 then
            local chosenGroup = availableGroups[math.random(#availableGroups)]
            currentTarget = {
                x = math.random() * (WIDTH * 0.75 - WIDTH * 0.25) + WIDTH * 0.25,
                y = math.random() * (HEIGHT * 0.75 - HEIGHT * 0.25) + HEIGHT * 0.25,
                radius = 50,
                colorGroup = chosenGroup,
                fillPct = 0
            }
        else
            gameState = "finished"
        end
    end
    if currentTarget then
        for i, person in ipairs(people) do
            if not person.corralled and person.colorGroup == currentTarget.colorGroup then
                local dx = person.x - currentTarget.x
                local dy = person.y - currentTarget.y
                local d = math.sqrt(dx * dx + dy * dy)
                if d < currentTarget.radius then
                    person.corralled = true
                    corralledCount[person.colorGroup] = (corralledCount[person.colorGroup] or 0) + 1
                end
            end
        end
        local total = initialCount[currentTarget.colorGroup] or 1
        local pct = (corralledCount[currentTarget.colorGroup] or 0) / total * 100
        currentTarget.fillPct = pct
        if pct >= 100 then
            currentTarget = nil
        end
    end
end

-- Draw the target circle.
function drawTarget()
    if currentTarget then
        local grp = currentTarget.colorGroup
        local col = groupColors[grp] or {r = 255, g = 255, b = 255}
        pushStyle()
        noFill()
        stroke(col.r, col.g, col.b)
        strokeWidth(3)
        ellipse(currentTarget.x, currentTarget.y, currentTarget.radius * 2)
        -- Fill based on percentage.
        local alpha = math.floor((currentTarget.fillPct / 100) * 255)
        fill(col.r, col.g, col.b, alpha)
        noStroke()
        ellipse(currentTarget.x, currentTarget.y, currentTarget.radius * 2)
        -- Draw percentage text.
        fill(255)
        fontSize(24)
        text(math.floor(currentTarget.fillPct) .. "%", currentTarget.x, currentTarget.y)
        popStyle()
    end
end

--------------------------------------------------
-- Agent Update & Draw Functions

function updatePerson(person, index)
    local shouldFlock = false
    if touchPos and activeColorGroup == person.colorGroup then
        if globalFlockDistance and globalFlockDistance <= maxFlockingDistance then
            shouldFlock = true
        end
    end
    
    if shouldFlock then
        -- For agents in the active color group, follow the dynamic target.
        person.randomImpulseApplied = false  -- reset impulse flag
        local targetX, targetY = calcCrowdTarget(person)
        local dx = targetX - person.x
        local dy = targetY - person.y
        local dist = math.sqrt(dx * dx + dy * dy)
        if dist > 1 then
            local reactionFactor = person.reactionFactor or 1.0
            local accel = 500 * reactionFactor  -- no extra scaling here now
            local ax = (dx / dist) * accel
            local ay = (dy / dist) * accel
            person.vx = person.vx + ax * DeltaTime
            person.vy = person.vy + ay * DeltaTime
        end
        lastTouchTime = ElapsedTime
    else
        -- Apply a continuous wandering force:
        local wanderAccel = 100 * (person.wanderFactor or 1.0) * 2
        local ax = (math.random() - 0.5) * wanderAccel
        local ay = (math.random() - 0.5) * wanderAccel
        person.vx = person.vx + ax * DeltaTime
        person.vy = person.vy + ay * DeltaTime

    end
    
    -- Apply damping.
    person.vx = person.vx * 0.95
    person.vy = person.vy * 0.95
    
    -- Update position.
    person.x = person.x + person.vx * DeltaTime
    person.y = person.y + person.vy * DeltaTime
    
    -- Side-wall collision: prevent agents from leaving the screen.
    if person.x < 0 then person.x = 0; person.vx = -person.vx end
    if person.x > WIDTH then person.x = WIDTH; person.vx = -person.vx end
    if person.y < 0 then person.y = 0; person.vy = -person.vy end
    if person.y > HEIGHT then person.y = HEIGHT; person.vy = -person.vy end
end

function drawDoodleWalker(person)
    if person.corralled then return end  -- Skip drawing corralled agents.
    pushMatrix()
    translate(person.x, person.y)
    lineCapMode(ROUND)
    local col = groupColors[person.colorGroup] or {r = 255, g = 255, b = 255}
    stroke(col.r, col.g, col.b)
    strokeWidth(2.5)
    local lineLength = 8
    local speed = math.sqrt(person.vx * person.vx + person.vy * person.vy)
    local baseAngle = speed > 0.1 and math.atan(person.vy, person.vx) or 0
    local freq = person.angleFreq or 1.5
    local rangeAngle = person.angleRange or 80
    local t = ElapsedTime
    
    local topOffset1 = math.sin(t * freq + (person.phaseTop or 0)) * rangeAngle
    local bottomOffset1 = math.sin(t * freq + (person.phaseBottom or 0)) * rangeAngle
    local effectiveAngleTop1 = baseAngle + topOffset1
    local effectiveAngleBottom1 = baseAngle + 180 + bottomOffset1
    local topX1 = math.cos(math.rad(effectiveAngleTop1)) * (lineLength / 2)
    local topY1 = math.sin(math.rad(effectiveAngleTop1)) * (lineLength / 2)
    local bottomX1 = math.cos(math.rad(effectiveAngleBottom1)) * (lineLength / 2)
    local bottomY1 = math.sin(math.rad(effectiveAngleBottom1)) * (lineLength / 2)
    line(0, 0, topX1, topY1)
    line(0, 0, bottomX1, bottomY1)
    
    local topOffset2 = math.sin(t * freq + (person.phaseTop2 or 0)) * rangeAngle
    local bottomOffset2 = math.sin(t * freq + (person.phaseBottom2 or 0)) * rangeAngle
    local effectiveAngleTop2 = baseAngle + topOffset2
    local effectiveAngleBottom2 = baseAngle + 180 + bottomOffset2
    local topX2 = math.cos(math.rad(effectiveAngleTop2)) * (lineLength / 2)
    local topY2 = math.sin(math.rad(effectiveAngleTop2)) * (lineLength / 2)
    local bottomX2 = math.cos(math.rad(effectiveAngleBottom2)) * (lineLength / 2)
    local bottomY2 = math.sin(math.rad(effectiveAngleBottom2)) * (lineLength / 2)
    line(0, 0, topX2, topY2)
    line(0, 0, bottomX2, bottomY2)
    
    fill(col.r, col.g, col.b)
    noStroke()
    ellipse(0, 0, 6.5)
    popMatrix()
end

--------------------------------------------------
-- Game Loop Functions

function setup()
    fps = 0
    lastTime = ElapsedTime
    resetPeople()
    updateGroupColors()
    resetCorralData()
end

function draw()
    background(40, 40, 50)
    globalFlockDistance = nil
    if touchPos and activeColorGroup then
        local minDist = math.huge
        for i, person in ipairs(people) do
            if person.colorGroup == activeColorGroup and not person.corralled then
                local dx = person.x - touchPos.x
                local dy = person.y - touchPos.y
                local d = math.sqrt(dx * dx + dy * dy)
                if d < minDist then
                    minDist = d
                end
            end
        end
        if minDist < math.huge then
            globalFlockDistance = minDist
        end
    end
    -- Check for slider changes and reset if needed.
    if numColorGroups ~= prevNumColorGroups then
        updateGroupColors()
        resetPeople()
        resetCorralData()
        prevNumColorGroups = numColorGroups
    end
    
    for i, person in ipairs(people) do
        updatePerson(person, i)
        drawDoodleWalker(person)
    end
    
    -- Update and draw the current target.
    updateTarget()
    drawTarget()
    
    -- If all groups have been corralled, show the reset button.
    if gameState == "finished" then
        local btn = drawResetButton()
    end
    
    wasTouching = (touchPos ~= nil)
end

-- Draw the reset button when the game is finished.
function drawResetButton()
    pushStyle()
    rectMode(CENTER)
    -- Set fill and stroke for the rounded button
    fill(0, 0, 0, 150)
    local btnX = WIDTH / 2
    local btnY = HEIGHT / 2
    local btnW = 150
    local btnH = 50
    -- Draw a rounded rectangle with a corner radius of 10 pixels
    drawRoundedRect(btnX, btnY, btnW, btnH, 10)
    fill(255)
    fontSize(24)
    font("STIXTwoText-Italic_Bold-Italic")
    text("Reset", btnX, btnY)
    popStyle()
    return { x = btnX - btnW/2, y = btnY - btnH/2, width = btnW, height = btnH }
end

--------------------------------------------------
-- Touch Handling

function touched(touch)
    if gameState == "finished" then
        -- If finished, check if reset button is pressed.
        if touch.state == BEGAN then
            local btn = drawResetButton()  -- Get button bounds.
            if touch.x >= btn.x and touch.x <= (btn.x + btn.width) and
            touch.y >= btn.y and touch.y <= (btn.y + btn.height) then
                resetGame()
            end
        end
        return
    end
    
    if touch.state == BEGAN then
        activeColorGroup = nil  -- Reset active color group at the beginning of a touch session.
    end
    if touch.state == BEGAN or touch.state == MOVING then
        touchPos = { x = touch.x, y = touch.y }
        lastTouchTime = ElapsedTime
        -- Use a dynamic threshold that increases for smaller groups.
        local threshold = 20 * math.sqrt(300 / numPeoplePerGroup)
        if threshold < 20 then threshold = 20 end
        if activeColorGroup == nil then
            for i, person in ipairs(people) do
                local dx = person.x - touch.x
                local dy = person.y - touch.y
                if math.sqrt(dx * dx + dy * dy) < threshold then
                    activeColorGroup = person.colorGroup
                    break
                end
            end
        end
    else
        touchPos = nil
        activeColorGroup = nil
    end
end

function drawRoundedRect(cx, cy, w, h, r)
    pushStyle()
    rectMode(CENTER)
    ellipseMode(CENTER)
    -- Draw corner ellipses
    ellipse(cx - w/2 + r, cy + h/2 - r, r*2)
    ellipse(cx + w/2 - r, cy + h/2 - r, r*2)
    ellipse(cx - w/2 + r, cy - h/2 + r, r*2)
    ellipse(cx + w/2 - r, cy - h/2 + r, r*2)
    -- Draw center rectangle
    rect(cx, cy, w - 2*r, h - 2*r)
    -- Draw top and bottom rectangles
    rect(cx, cy + (h - 2*r)/2, w - 2*r, 2*r)
    rect(cx, cy - (h - 2*r)/2, w - 2*r, 2*r)
    -- Draw left and right rectangles
    rect(cx - (w - 2*r)/2, cy, 2*r, h - 2*r)
    rect(cx + (w - 2*r)/2, cy, 2*r, h - 2*r)
    popStyle()
end

